"""
Also see 

https://pypi.python.org/pypi/RangeParser
(parse ranges like 10,20,30-50)
https://pypi.python.org/pypi/DateRangeParser
(parse ranges like 10-17 January, March 2014 - May 2015)
http://pythonhosted.org//ranger/tutorial/ranges.html#using-range-objects
(operations on ranges: intersection, span, difference, contains, contains all, 
overlaps, union etc. w/ both continuous and discrete [stepped] ranges)
(additionally, using these basic operations, it makes available a 
range map [any key between x-y => value] and range bucket [any key between
x-y => any corresponding values])
https://pypi.python.org/pypi/multirange
(return the gaps in ranges, a.k.a. the inverse of a range of discrete values)
https://github.com/brianhicks/cowboy
(uses operator overloading to enable `value in range` comparisons)
https://pypi.python.org/pypi/lrange/0.0.5
(lazy infinite ranges, like xrange)
https://docs.python.org/3/library/collections.abc.html#collections.abc.Sequence
(abstract base class for sequences)
https://github.com/zacharyvoase/daterange/
(date ranges)
https://pypi.python.org/pypi/datetimes/
(date ranges using timedelta for intervals)
"""

def pad(l, n, value):
    l = list(l)
    while len(l) < n:
        l.append(value)
    return l

# TODO: use `ranges` under the hood, so we can do more kinds of ranges
# TODO: support ranges that combine , and .., e.g. `3,4,5..10..2`
# to produce `3,4,5,7,9`
# TODO: distinguish between closed ranges (..) and open ranges (...)
def parse(description, 
    sequence_delimiters=[','], 
    closed_range_delimiters=['..', ' - '], 
    open_range_delimiters=['...', ' -- '], 
    step_delimiters=['..', '/']):

    if ',' in description:
        return description.split(',')
    elif '..' in description:
        params = description.split('..')
        start, stop, step = map(int, pad(params, n=3, value=1))
        return range(start, stop, step)
    else:
        raise ValueError()



import datetime
import re
import string
from collections import OrderedDict

# Python 2/3 compatibility
try:
    basestring = basestring
    unicode = unicode
except NameError:
    basestring = str
    unicode = str

try:
    import __builtin__ as builtins
except ImportError:
    import builtins


class Step(object):
    @classmethod
    def cast(self, raw):
        for cls in [Integer, Float, Character, Date]:
            if cls.match(raw):
                return cls(raw)
        return raw

    def __init__(self, raw):
        self.raw = raw


class Integer(Step):
    def __init__(self, raw, width=None):
        self.raw = raw

        if isinstance(raw, basestring):
            self.value = int(raw)
            if width is None and raw.startswith('0'):
                width = len(raw)
        else:
            self.value = raw

        self.width = width or 0

    @classmethod
    def match(cls, value):
        is_int = isinstance(value, int)
        is_numeric = isinstance(value, basestring) and re.match(r'^[0-9]+$', value)
        return is_int or is_numeric

    def to(self, stop, step):
        r = builtins.range(self.value, stop.value, step.value) + [stop]
        return [Integer(v, self.width) for v in r]

    def __unicode__(self):
        return unicode(self.value).zfill(self.width)


class Float(Integer):
    def __init__(self, raw, width=None, decimals=None):
        self.raw = raw
        if isinstance(raw, basestring):
            self.value = float(raw)
            if width is None:
                if raw.startswith('0') or raw.endswith('0'):
                    width = len(raw)
                    decimals = len(raw.split('.')[1])
        else:
            self.value = raw

        self.width = width or 0
        self.decimals = decimals

    @classmethod
    def match(cls, value):
        is_float = isinstance(value, float)
        is_numeric = isinstance(value, basestring) and re.match(r'^[0-9]+\.[0-9]+$', value)
        return is_float or is_numeric

    def to(self, other, step):
        r = [self]
        while r[-1].value < other.value:
            v = Float(r[-1].value + step.value, self.width, self.decimals)
            r.append(v)

        return r

    def __unicode__(self):
        right = self.decimals

        if right is None:
            value = self.value
        else:
            value = round(self.value, right)

        integer, fractional = unicode(value).split('.')
        fractional = fractional.ljust(right, '0')
        value = ".".join((integer, fractional))
        return value.zfill(self.width)


class Character(Step):
    def __init__(self, raw):
        self.raw = self.value = raw
        if raw.islower():
            self.case = string.lowercase
        else:
            self.case = string.uppercase

    @classmethod
    def match(cls, value):
        return isinstance(value, basestring) and re.match(r'^[a-zA-Z]{1}$', value)

    @property
    def ix(self):
        return self.case.index(self.value)

    def to(self, other, step):
        r = range(self.ix, other.ix, step) + [other.ix]
        return [Character(self.case[ix]) for ix in r]

    def __unicode__(self):
        return unicode(self.value)


class Date(Step):
    """
    We're going to do this without dateutil, because
    we need to preserve the initial format so we can
    output a formatted date at the end of the ride.

    Supported formats:

    - YYYY
    - YYYY-MM
    - YYYY-MM-DD
    - MM-DD

    - HH:MM

    - combination of the above

    (Only years, months etc. are not supported:
    these would count as regular integer-based ranges.
    Shorthand years are not supported either: these
    would be ambigous and could be interpreted as months.)

    Supported separators are dash, dot, underscore
        - . _
    """

    multipliers = {
        'S': 1,
        'M': 60,
        'H': 60 * 60,
        'D': 60 * 60 * 24,
        'W': 60 * 60 * 24 * 7,
        'Y': 60 * 60 * 24 * 365,
    }

    components = OrderedDict((
        ('year', r"([0-9]{4})"),
        ('ym', r"([\-\.\_\ \\])"),
        ('month', r"([0-9]{2})"),
        ('md', r"([\-\.\_\ \\])"),
        ('day', r"([0-9]{2})"),
        ('dt', r"([\-\.\_\ \\])"),
        ('hour', r"([0-9]{2})"),
        ('hm', r"([Hh\:\-\.\_\ \\])"),
        ('minute', r"([0-9]{2})"),
        ('ms', r"([Mm])"),
        ('second', r"([0-9]{2})"),
        ('s', r"([Ss])"),
    ))

    formats = OrderedDict((
        ('year', '%Y'),
        ('month', '%m'),
        ('day', '%d'),
        ('hour', '%H'),
        ('minute', '%M'),
        ('second', '%s'),
    ))

    pattern = r"""
        ^
        ({year}{ym}?{month}{md}?{day}|{year}{ym}?{month}|{month}{md}?{day})?
        {dt}?
        ({hour}{hm}?{minute}{ms}?({second}{s})?)?
        $
        """.format(**components)

    pattern = re.compile(pattern, re.VERBOSE)

    # Python doesn't allow repeating group names (even if mutually exclusive)
    # so we have to resort to this kind of silliness to make sense of our RegEx.
    groups = [
        'year', 'ym', 'month', 'md', 'day',
        'year', 'ym', 'month',
        'month', 'md', 'day',
        'dt',
        '_time',
        'hour', 'hm', 'minute', 'ms', '_second', 'second', 's'
        ]


    @classmethod
    def match(cls, value):
        if cls._is_date(value):
            return True
        elif isinstance(value, basestring) and re.match(cls.pattern, value):
            return True
        else:
            return False

    @classmethod
    def _is_date(cls, value):
        d = isinstance(value, datetime.date)
        t = isinstance(value, datetime.time)
        dt = isinstance(value, datetime.datetime)
        return d or t or dt

    def __init__(self, raw, format=None):
        self.raw = raw

        if format:
            self._format = format
        else:
            matches = list(re.match(self.pattern, raw).groups()[1:])
            tuples = zip(self.groups, matches)
            self._format = dict([(k, v) for k, v in tuples if v])

        if self._is_date(raw):
            self.value = raw
        else:
            # extract date and time
            self.value = self._parse_date(self._format)

    def _parse_date(self, format):
        year = int(format.get('year', 9999))
        month = int(format.get('month', 1))
        day = int(format.get('day', 1))
        hour = int(format.get('hour', 0))
        minute = int(format.get('minute', 0))
        second = int(format.get('second', 0))

        if 'month' in format and '_time' in format:
            return datetime.datetime(year, month, day, hour, minute, second)
        elif 'month' in format:
            return datetime.datetime(year, month, day)
        elif '_time' in format:
            return datetime.datetime(9999, 1, 1, hour, minute, second)
        else:
            raise ValueError()

    def _parse_step(self, step):
        if isinstance(step, basestring):
            magnitude = step[:-1]
            unit = step[-1:].upper()
            step = float(magnitude) * self.multipliers[unit]

        return datetime.timedelta(seconds=step)

    @property
    def format(self):
        date_components = self.formats.keys()
        components = [c for c in self.components]
        first = next(c for c in date_components if c in self._format)
        last = next(c for c in reversed(date_components) if c in self._format)
        start = components.index(first)
        stop = components.index(last) + 1

        fmt = ''
        for component in components[start:stop]:
            # replace date components with strftime placeholders
            if component in date_components and component in self._format:
                fmt += self.formats[component]
            # keep the original separators intact
            else:
                fmt += self._format.get(component, '')

        return fmt

    def to(self, other, step):
        step = self._parse_step(step)

        r = [self]
        while r[-1].value < other.value:
            r.append(Date(r[-1].value + step, self._format))

        return r

    def __unicode__(self):
        return self.value.strftime(self.format)



from dateutil.relativedelta import relativedelta
from funcy import pairwise


def interval(start=None, stop=None, **offsets):
    delta = relativedelta(**offsets)

    if start:
        return (start, start + delta)
    elif stop:
        return (stop - delta, stop)
    else:
        raise ValueError()

def range(start, stop, align=False, alignment='left', closed=False, intervals=None, **offsets):
    """
    * days, weeks, months and years interval
    * align means you "snap" the daterange to the nearest start of the week/month/year etc.;
      with alignment, you decide whether to snap to the next valid occurence, or the previous one.
      e.g. `align="day"`
    * intervals is similar to step, but returns (start, stop) pairs for every step
    * a closed interval includes both start and stop
    """

    # TODO: incorporate alignment
    steps = []
    last = start

    delta = relativedelta(**offsets)

    while last < stop:
        yield last
        last = last + delta

    if closed:
        yield last



from .types import Step
from .parse import parse
from . import utils
from . import date


def values(r, string=False):
    if string:
        return [utils.unicode(v) for v in r]
    else:
        return [v.value for v in r]


# calculates stop from start, or vice versa, no intermediate steps,
# only for dates, letters and numbers
def interval(start, **options):
    raise NotImplementedError()


# a space is like a range, but with a number of steps (based on which the
# appropriate step size is then constructed) rather than a step size
def space(start, stop, num):
    return range()


# TODO: `step` for floats and dates, if not defined, should be
# dynamic based on the granularity of `start`
# e.g. 2013-03 ==> +1M
# e.g. 1.23 ==> +0.01
# e.g. 2H ==> +1H
#
# TODO: `n` argument, so e.g. you specify start and n=10
# and it will increment 9 times
def range(start, stop, step=None, closed=False, string=False, intervals=False, **options):
    """
    This function is (or should be) a strict superset of the `range`
    builtin and so can be safely loaded as `from ranges import range`.

    Negative steps should be supported too.

    Steps can be numbers (translates to skips in the range) or
    strings (e.g. 'minute', 'day', ...)
    """

    start = Step.cast(start)
    stop = Step.cast(stop)
    step = Step.cast(step)

    if start.__class__.__name__ != stop.__class__.__name__:
        raise ValueError()

    # a stepless range can be useful if we simply wish to
    # cast the start and stop values from strings into
    # a more appropriate type
    if step is False:
        return values([start, stop], string)

    r = start.to(stop, step)

    # all Step#to methods should return a closed range,
    # so if an open range is demanded, we just chop
    # off the last value
    if not closed:
        r = r[:-1]

    # we can return the range's values in string format
    # we were given, or cast into the proper types
    r = values(r, string)

    # intervals are pairs, e.g. (a, b) (b, c) (c, d)
    if intervals:
        r = utils.intervals(r)

    return r



import unittest
import ranges


def TestRange(unittest.TestCase):
    """ Test ranges of simple Python data types. """

    def test_int(self):
        """ We can generate a range of integers. """

        """
        >>> range(0, 10, 3)
        [0, 3, 6, 9]
        """

        raise NotImplementedError()

    def test_closed_int(self):
        """ We can generate a closed range of integers. """
        raise NotImplementedError()

    def test_float(self):
        """ We can generate a range of floats. """

        """
        >>> range(9.2, 10.0, 0.25)
        [9.2, 9.45, 9.7, 9.95]
        """

        raise NotImplementedError()

    def test_closed_float(self):
        """ We can generate a closed range of floats. """
        raise NotImplementedError()

    def test_character(self):
        """ We can generate a range of characters. """
        raise NotImplementedError()

    def test_closed_characters(self):
        """ We can generate a closed range of characters. """
        raise NotImplementedError()

    def test_date(self):
        """ We can generate a range of dates. """

        """
        >>> range(date(2013,1,1), date(2013,1,3))
        ...
        """

        raise NotImplementedError()

    def test_closed_date(self):
        """ We can generate a closed range of dates. """
        raise NotImplementedError()

    def test_time(self):
        """ We can generate a range of times. """
        raise NotImplementedError()

    def test_closed_time(self):
        """ We can generate a closed range of times. """
        raise NotImplementedError()

    def test_datetime(self):
        """ We can generate a range of datetimes. """
        raise NotImplementedError()

    def test_closed_datetime(self):
        """ We can generate a closed range of datetimes. """
        raise NotImplementedError()


def TestHRange(unittest.TestCase):
    """ Test complex ranges based on string definitions. """

    def test_int(self):
        """ We can generate a range of integers from strings. """
        raise NotImplementedError()

    def test_padded_int(self):
        """ We can generate a range of padded integers from strings. """

        """
        >>> range('0000', '0500', 100)
        ['0000', '0100', '0200', '0300', '0400']
        """

        raise NotImplementedError()

    def test_closed_padded_int(self):
        """ We can generate a closed range of padded integers from strings. """
        raise NotImplementedError()

    def test_float(self):
        """ We can generate a range of floats from strings. """
        raise NotImplementedError()

    def test_padded_float(self):
        """ We can generate a range of padded floats from strings. """
        raise NotImplementedError()

    def test_rounded_float(self):
        """ We can generate a range of rounded floats from strings. """
        raise NotImplementedError()

    def test_padded_rounded_float(self):
        """ We can generate a range of padded and rounded floats from strings. """
        raise NotImplementedError()

    def test_closed_padded_rounded_float(self):
        """ We can generate a closed range of padded and rounded floats from strings. """
        raise NotImplementedError()

    def test_date(self):
        """ We can generate a range of dates that respect the original string formatting. """

        """
        >>> range('2013-02.01', '2013-02.03')
        ['2013-02.01', '2013-02.02', '2013-02.03']
        """

        raise NotImplementedError()

    def test_closed_date(self):
        """ We can generate a closed range of dates from strings. """
        raise NotImplementedError()

    def test_time(self):
        """ We can generate a range of times that respect the original string formatting. """

        """
        >>> range('10H13M', '10H17M', step=2, closed=True)
        ['10H13M', '10H15M', '10H17M']
        """

        raise NotImplementedError()

    def test_closed_time(self):
        """ We can generate a closed range of times from strings. """
        raise NotImplementedError()

    def test_datetime(self):
        """ We can generate a range of datetimes that respect the original string formatting. """
        raise NotImplementedError()

    def test_closed_date(self):
        """ We can generate a closed range of datetimes from strings. """
        raise NotImplementedError()




# tests, until we get a setup.py sorted out etc.
if __name__ == '__main__':
    print range(0, 10, 2)
    print range('00', '10', 2, string=True)
    print range('1.100', '1.650', 0.175, string=False)
    print range('1.100', '1.650', 0.175, string=True)
    print range('A', 'H', 1)
    print range('a', 'H', 2)
    print range('2013-12 11:00', '2013-12 11:21', '3M', string=True)
    print range('2013-12-31 23:55', '2014-01-01 00:10', '10M', string=True)



from setuptools import find_packages, setup

setup(name='formatted-ranges',
      description='Generate various kinds of ranges, either in whatever Python \
            data type is appropriate or as a list of strings that retain all   \
            the various quirks of your start, stop and step arguments.',
      #long_description=open('README.md').read(),
      author='Stijn Debrouwere',
      author_email='stijn@debrouwere.org',
      url='http://stdbrouw.github.com/python-ranges/',
      download_url='http://www.github.com/stdbrouw/python-ranges/tarball/master',
      version='0.2.1',
      license='MIT',
      packages=find_packages(),
      keywords='data range utilities',
      classifiers=['Development Status :: 3 - Alpha',
                   'Intended Audience :: Developers',
                   'Intended Audience :: Science/Research',
                   'License :: OSI Approved :: MIT License',
                   'Operating System :: OS Independent',
                   'Programming Language :: Python',
                   'Topic :: Text Processing',
                   'Topic :: Utilities'],
      )
